@@ -60,10 +60,6 @@ GRL_LOG_DOMAIN(flickr_log_domain);
 #define PERSONAL_SOURCE_NAME "%s's Flickr"
 #define PERSONAL_SOURCE_DESC "A source for browsing and searching %s' flickr photos"
 
-#define AUTHOR      "Igalia S.L."
-#define LICENSE     "LGPL"
-#define SITE        "http://www.igalia.com"
-
 typedef struct {
   GrlMediaSource *source;
   GrlMediaSourceResultCb callback;
@@ -244,6 +240,7 @@ grl_flickr_source_finalize (GObject *object)
 
   source = GRL_FLICKR_SOURCE (object);
   g_free (source->priv->user_id);
+  g_object_unref (source->priv->flickr);
 
   G_OBJECT_CLASS (grl_flickr_source_parent_class)->finalize (object);
 }
@@ -380,7 +377,7 @@ getInfo_cb (GFlickr *f, GHashTable *photo, gpointer user_data)
     update_media (ms->media, photo);
   }
 
-  ms->callback (ms->source, ms->media, ms->user_data, NULL);
+  ms->callback (ms->source, ms->metadata_id, ms->media, ms->user_data, NULL);
 }
 
 static void
@@ -702,7 +699,7 @@ grl_flickr_source_metadata (GrlMediaSource *source,
   const gchar *id;
 
   if (!ms->media || (id = grl_media_get_id (ms->media)) == NULL) {
-    ms->callback (ms->source, ms->media, ms->user_data, NULL);
+    ms->callback (ms->source, ms->metadata_id, ms->media, ms->user_data, NULL);
     return;
   }
 
